from multiprocessing import Process, Manager, sharedctypes
from collections import defaultdict
import time

from ..utils import printProgressBar
from ..utils import Timeout


class WorkerProcess(Process):
    
    def __init__(self, shared_results_dict, shared_status_dict, shared_error_dict, task, timeout):
        super(WorkerProcess, self).__init__()
    
        self.shared_results_dict = shared_results_dict
        self.shared_status_dict = shared_status_dict
        self.shared_error_dict = shared_error_dict
        self.task = task
        self.timeout = timeout
        
    def run(self):
        """
        We need a way to surface the failure reasons.
        In case of actual failed task.
        """
        
        task_idx, func, args, kwargs = self.task
        
        try:

            with Timeout(seconds=self.timeout):

                result = func(*args, **kwargs)
                self.shared_error_dict[task_idx] = None
                self.shared_status_dict[task_idx] = 'success'
                self.shared_results_dict[task_idx] = result

        except Exception as e:

            self.shared_error_dict[task_idx] = str(e)
            self.shared_status_dict[task_idx] = 'failure'
            self.shared_results_dict[task_idx] = None 


class ManagedProcesses():
    
    def __init__(self, n_procs=4, verbose=False, show_progress=True, timeout=60*60, *args, **kwargs):
        
        self.task_idx = 0
        
        self.tasks = []
        self.pending_tasks = []
        
        self.failed_tasks_idx = []
        self.successful_tasks_idx = []
        self.task_idx_attempt_dict = defaultdict(int)
        
        manager = Manager()
        self.shared_results_dict = manager.dict()
        self.shared_status_dict = manager.dict()
        self.shared_error_dict = manager.dict()
        
        self.subprocs = dict()
        self.subprocs_start_times = dict()
        
        self.n_procs = n_procs
        self.verbose = verbose
        self.show_progress = show_progress
        self.timeout = timeout
        
        self.scattered_vars = dict()
        
    def globalize(self, local_var, global_name):
        
        self.scattered_vars[global_name] = local_var
        
    def submit(self, func, *args, **kwargs):
        
        if len(self.scattered_vars) > 0:
            func.__globals__.update(self.scattered_vars)
        
        task_idx = int(self.task_idx)
        self.tasks.append([task_idx, func, args, kwargs])
        
        self.task_idx += 1
        
    def execute_task(self, task):

        timeout = int(self.timeout)
        
        subproc = WorkerProcess(
            self.shared_results_dict, 
            self.shared_status_dict, 
            self.shared_error_dict,
            task,
            timeout)

        subproc.daemon = True
        subproc.start()
        return subproc

    def results(self):
        
        self.tasks = self.tasks[::-1]
        num_tasks = len(self.tasks)
        
        if(self.show_progress):
            printProgressBar(0, num_tasks)
        
        while(len(self.pending_tasks) > 0 or len(self.tasks) > 0):

            # if we can run more subprocs and if there are tasks remaining
            # run a task in subproc
            while(len(self.subprocs) < self.n_procs and len(self.tasks) > 0):

                task = self.tasks.pop()
                task_idx = task[0]
                
                self.pending_tasks.append(task)
                
                subproc = self.execute_task(task)
                self.task_idx_attempt_dict[task_idx] += 1

                self.subprocs[task_idx] = subproc
                self.subprocs_start_times[task_idx] = time.time()
                
            # these are not just pending but completed tasks, whether failed or succeeded
            if(len(self.shared_status_dict.keys()) > 0):

                for tried_task_idx in self.shared_status_dict.keys():

                    if self.shared_status_dict[tried_task_idx] == 'success':
        
                        if self.verbose:
                            print('Task execution successful: {}\n'.format(tried_task_idx))
                        self.successful_tasks_idx.append(tried_task_idx)
                        
                    elif self.shared_status_dict[tried_task_idx] == 'failure':

                        if self.verbose:
                            print('Task execution failed: {}\n'.format(tried_task_idx))

                        # we need to push these back to tasks
                        # since we need to start the process again, 
                        # no need for any wait time 
                        # (there is nothing to wait for! it failed!)
                        self.failed_tasks_idx.append(tried_task_idx)
                        
                    self.shared_status_dict.pop(tried_task_idx)
                        
            if(len(self.pending_tasks) > 0):
                
                for _ in range(len(self.pending_tasks)):
                    
                    pending_task = self.pending_tasks.pop()
                    pending_task_idx = pending_task[0]
                    
                    current_time = time.time()
                    start_time = self.subprocs_start_times[pending_task_idx]
                    passed_time = current_time - start_time
                    
                    if self.verbose:
                        print('Looking at pending task {}'.format(pending_task))
                    
                    if pending_task_idx in self.failed_tasks_idx:
                        
                        fail_count = self.task_idx_attempt_dict[task_idx]
                        
                        if fail_count < 3:
                            
                            if self.verbose:
                                print('    pending --> tasks: {}'.format(pending_task_idx))
                                print('    failed {} times'.format(fail_count))
                                
                            # don't append since that will try to do this again right away
                            self.tasks.insert(0, pending_task)
                            self.failed_tasks_idx.remove(pending_task_idx)  # give it another chance
                                
                        else:
                            
                            if self.verbose:
                                print('    pending --> failed tasks: {}'.format(pending_task_idx))
                                print('    * reached MAX_ATTEMPT\n')
                                
                            # just leave the failed idx alone
                            
                    elif pending_task_idx in self.successful_tasks_idx:

                        if self.verbose:
                            print('    pending --> successful tasks: {}\n'.format(pending_task_idx))
                    
                    # premature and sudden death of Process
                    elif pending_task_idx not in self.subprocs:
                        
                        if self.verbose:
                            print('    Process terminated abruptly while task {} pending'.format(pending_task_idx))
                            print('    pending --> tasks')
                        
                        fail_count = self.task_idx_attempt_dict[task_idx]
                        
                        if fail_count < 3:
                            
                            self.tasks.insert(0, pending_task)
                            
                        else:
                            
                            self.failed_tasks_idx.append(pending_task_idx)
                    
                    # timeout 
                    elif passed_time > self.timeout:
                        
                        self.subprocs[pending_task_idx].terminate()
                        
                        if self.verbose:
                            print('    pending task {} --> timed out'.format(pending_task_idx))
                    
                        fail_count = self.task_idx_attempt_dict[task_idx]
                        
                        if fail_count < 3:
                            
                            self.tasks.insert(0, pending_task)
                            
                        else:
                            
                            self.failed_tasks_idx.append(pending_task_idx)
                        
                    else:
                        
                        if self.verbose:
                            print('    nothing to update\n')

                        self.pending_tasks.insert(0, pending_task)
                        
            if(self.show_progress):

                cur_len = int(len(self.successful_tasks_idx) + len(self.failed_tasks_idx))  
                printProgressBar(cur_len, num_tasks)
            
            # subproc dies only after updating shared dicts
            # by the time the subprocs list updated and IF one drops
            # that means the shared dicts HAVE to have been updated
            self.subprocs = {k: v for k, v in self.subprocs.items() if v.is_alive()}

            if self.verbose:
                print('SUBPROC STATUS')
                print('number of outstanding tasks: {} ( {} )'.format(len(self.tasks), self.tasks))
                print('number of running procs: {}'.format(len(self.subprocs))); print()

            # if all procs are busy, there is no gain from trying to 
            # assign work to them at the moment
            if(len(self.subprocs) == self.n_procs):
                time.sleep(0.1)

            # if number of tasks is 0, there is no gain from trying to
            # assign work to them at the moment
            # also, if subproc is still running, sleep to give it time to finish
            if(len(self.subprocs) > 0 and len(self.tasks) == 0):
                time.sleep(0.1)
                
        if(self.show_progress):
            printProgressBar(num_tasks, num_tasks)
            print()
            
        return self.shared_results_dict.values()
                
                


from ..utils import printProgressBar

from concurrent.futures import ProcessPoolExecutor
import concurrent.futures


class ConcurrentFuturesProcessPool():
    
    def __init__(self, n_procs=4, *args, **kwargs):
        
        self.executor = executor = ProcessPoolExecutor(max_workers=n_procs)
        self.futures_list = []
        self.results_list = []
    
    def submit(self, func, *args, **kwargs):
        
        future = self.executor.submit(func, *args, **kwargs)
        self.futures_list.append(future)
        
    def results(self):
        
        num_tasks = len(self.futures_list)
        
        if num_tasks > 0:
            printProgressBar(0, num_tasks)

            for idx, future in enumerate(concurrent.futures.as_completed(self.futures_list)):

                self.results_list.append(future.result())
                printProgressBar(idx, num_tasks)

            self.futures_list = []

            printProgressBar(num_tasks, num_tasks)

            self.executor.shutdown(wait=True)
        
        return self.results_list
    
    


from setuptools import setup, find_packages


import ezFutures

VERSION = ezFutures.__version__

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
	name="ezFutures", 
	version=VERSION,
	author="Jay Kim",
	description="Parallism made easy (and safe) with futures API",
	long_description=long_description,
	long_description_content_type="text/x-rst",
	url="https://github.com/mozjay0619/ezFutures",
	license="DSB 3-clause",
	packages=find_packages(),
	#install_requires=["numpy>=1.18.2", "pandas>=0.25.3", "psutil>=5.7.0"]
	)



from .core.managed_processes import ManagedProcesses
from .core.concurrent_futures_process_pool import ConcurrentFuturesProcessPool

class ezFutures():
    
    def __init__(self, n_procs=4, verbose=False, show_progress=True, timeout=60*60, parallelism_mode='managed.processes'):
        
        self.parallelism_mode = parallelism_mode

        if self.parallelism_mode=='managed.processes':
            self.core = ManagedProcesses(n_procs=n_procs, verbose=verbose, show_progress=show_progress, timeout=timeout)
        elif self.parallelism_mode=='concurrent.futures.process.pool':
            self.core = ConcurrentFuturesProcessPool(n_procs=n_procs)
    
    def submit(self, func, *args, **kwargs):
        
        self.core.submit(func, *args, **kwargs)

    def globalize(self, local_var, global_name):

        if self.parallelism_mode=='managed.processes':
            return self.core.globalize(local_var, global_name)
        else:    
            raise Exception('[globalize] method only supported for parallelism_mode: managed.processes')
        
    def results(self):
        
        return self.core.results()

    def errors(self):

        if self.parallelism_mode=='managed.processes':
            return self.core.shared_error_dict.items()
        else:    
            raise Exception('[errors] method only supported for parallelism_mode: managed.processes')




import signal

# https://stackoverflow.com/questions/2281850/timeout-function-if-it-takes-too-long-to-finish
class Timeout:
    def __init__(self, seconds=1, error_message='Timeout'):
        self.seconds = seconds
        self.error_message = error_message
    def handle_timeout(self, signum, frame):
        raise TimeoutError(self.error_message)
    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)
    def __exit__(self, type, value, traceback):
        signal.alarm(0)


def printProgressBar(iteration, total, prefix = 'Progress:', suffix = '', decimals = 1, length = 50, fill = '█', printEnd = "\r"):
    """
    Call in a loop to create terminal progress bar
    @params:
        iteration   - Required  : current iteration (Int)
        total      - Required  : total iterations (Int)
        prefix    - Optional  : prefix string (Str)
        suffix    - Optional  : suffix string (Str)
        decimals    - Optional  : positive number of decimals in percent complete (Int)
        length    - Optional  : character length of bar (Int)
        fill        - Optional  : bar fill character (Str)
        printEnd    - Optional  : end character (e.g. "\r", "\r\n") (Str)
    """
    percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
    filledLength = int(length * iteration // total)
    bar = fill * filledLength + '-' * (length - filledLength)
    print(f'\r{prefix} |{bar}| {percent}% {suffix}', end = printEnd)
    # Print New Line on Complete
    if iteration == total: 
        print(f'\r', end = printEnd)